# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the GNU Affero General Public License 3.0 (the "License").
# A copy of the License may be obtained with this software package or at
#
#      https://www.gnu.org/licenses/agpl-3.0.en.html
#
# Use of this file is prohibited except in compliance with the License. Any
# modifications or derivative works of this file must retain this copyright
# notice, and modified files must contain a notice indicating that they have
# been altered from the originals.
#
# Covalent is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the License for more details.
#
# Relief from the License may be granted by purchasing a commercial license.

from contextlib import contextmanager
from os import path
from pathlib import Path
from typing import BinaryIO, Dict, Generator, List, Optional, Union

from sqlalchemy import create_engine
from sqlalchemy.orm import Session, sessionmaker

from alembic import command
from alembic.config import Config

from .._shared_files.config import get_config
from . import models
from .storage_backends import LocalStorageBackend, StorageBackend

default_backend_map = {"local": LocalStorageBackend(base_dir=get_config("workflow_data.base_dir"))}


class DataStore:
    def __init__(
        self,
        db_URL: Optional[str] = None,
        storage_backend_map: Dict[str, StorageBackend] = default_backend_map,
        initialize_db: bool = False,
        **kwargs,
    ):
        if db_URL:
            self.db_URL = db_URL
        else:
            self.db_URL = "sqlite+pysqlite:///" + get_config("workflow_data.db_path")

        self.storage_backend_map = storage_backend_map
        self.engine = create_engine(self.db_URL, **kwargs)
        self.Session = sessionmaker(self.engine)

        # flag should only be used in pytest - tables should be generated using migrations
        if initialize_db:
            models.Base.metadata.create_all(self.engine)

    def run_migrations(self):

        alembic_ini_path = Path(path.join(__file__, "./../../../alembic.ini")).resolve()
        alembic_config = Config(alembic_ini_path)
        alembic_config.attributes["configure_logger"] = False
        alembic_config.set_main_option("sqlalchemy.url", self.db_URL)
        command.upgrade(alembic_config, "head")

    @contextmanager
    def session(self) -> Generator[Session, None, None]:
        with self.Session.begin() as session:
            yield session


class DevDataStore(DataStore):
    def __init__(
        self,
        db_URL: Optional[str] = None,
        initialize_db: bool = False,
        storage_backend_map: Dict[str, StorageBackend] = default_backend_map,
        **kwargs,
    ):
        if not db_URL:
            db_path = get_config("user_interface.dispatch_db")
            sqlite = ".sqlite"
            db_URL = f"sqlite+pysqlite:///{db_path.split(sqlite)[0]}_dev{sqlite}"
        super().__init__(db_URL, storage_backend_map, initialize_db, **kwargs)


class DataStoreSession:
    def __init__(self, session: Session, metadata={}):
        self.db_session = session
        self.metadata = metadata
        self.pending_uploads = []
        self.pending_deletes = []

    def queue_upload(self, data: BinaryIO, storage_type: str, storage_path: str, file_name: str):
        self.pending_uploads.append((data, storage_type, storage_path, file_name))

    def queue_delete(self, storage_type: str, storage_path: str, file_name: str):
        self.pending_deletes.append((storage_type, storage_path, file_name))


class DataStoreNotInitializedError(Exception):
    """Exception raised when a database action is attempted before the database is initialized."""

    def __init__(self, message="Database is not initialized."):
        self.message = message
        super().__init__(self.message)


# we can switch this to any class instance that has a db_URL property that points to the db
# which we want to run migrations against - this command also creates the db without tables
# via create_engine()
workflow_db = DevDataStore(echo=True)



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the GNU Affero General Public License 3.0 (the "License").
# A copy of the License may be obtained with this software package or at
#
#      https://www.gnu.org/licenses/agpl-3.0.en.html
#
# Use of this file is prohibited except in compliance with the License. Any
# modifications or derivative works of this file must retain this copyright
# notice, and modified files must contain a notice indicating that they have
# been altered from the originals.
#
# Covalent is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the License for more details.
#
# Relief from the License may be granted by purchasing a commercial license.

import codecs
from concurrent.futures import ThreadPoolExecutor

import cloudpickle as pickle
from flask import Blueprint, Response, jsonify, make_response, request
from sqlalchemy.orm import Session

import covalent_dispatcher as dispatcher
from covalent._data_store.models import Lattice
from covalent._results_manager.result import Result
from covalent._results_manager.results_manager import result_from
from covalent._results_manager.write_result_to_db import MissingLatticeRecordError

from .._db.dispatchdb import DispatchDB

bp = Blueprint("dispatcher", __name__, url_prefix="/api")


@bp.before_app_first_request
def start_pools():
    global workflow_pool
    global tasks_pool

    workflow_pool = ThreadPoolExecutor()
    tasks_pool = ThreadPoolExecutor()


@bp.route("/submit", methods=["POST"])
def submit() -> Response:
    """
    Function to accept the submit request of
    new dispatch and return the dispatch id
    back to the client.

    Args:
        None

    Returns:
        dispatch_id: The dispatch id in a json format
                     returned as a Flask Response object.
    """

    json_lattice = request.get_data()

    dispatch_id = dispatcher.run_dispatcher(json_lattice, workflow_pool, tasks_pool)

    return jsonify(dispatch_id)


@bp.route("/cancel", methods=["POST"])
def cancel() -> Response:
    """
    Function to accept the cancel request of
    a dispatch.

    Args:
        None

    Returns:
        Flask Response object confirming that the dispatch
        has been cancelled.
    """
    dispatch_id = request.get_data().decode("utf-8")

    dispatcher.cancel_running_dispatch(dispatch_id)

    return jsonify(f"Dispatch {dispatch_id} cancelled.")


@bp.route("/db-path", methods=["GET"])
def db_path() -> Response:
    db_path = DispatchDB()._dbpath
    return jsonify(db_path)


@bp.route("/result/<dispatch_id>", methods=["GET"])
def get_result(dispatch_id) -> Response:
    args = request.args
    wait = args.get("wait", default=False, type=lambda v: v.lower() == "true")
    status_only = args.get("status_only", default=False, type=lambda v: v.lower() == "true")
    while True:
        with Session(DispatchDB()._get_data_store().engine) as session:
            lattice_record = (
                session.query(Lattice).where(Lattice.dispatch_id == dispatch_id).first()
            )
        try:
            if not lattice_record:
                return (
                    jsonify(
                        {"message": f"The requested dispatch ID {dispatch_id} was not found."}
                    ),
                    404,
                )
            elif not wait or lattice_record.status in [
                str(Result.COMPLETED),
                str(Result.FAILED),
                str(Result.CANCELLED),
                str(Result.POSTPROCESSING_FAILED),
                str(Result.PENDING_POSTPROCESSING),
            ]:
                output = {
                    "id": dispatch_id,
                    "status": lattice_record.status,
                }
                if not status_only:
                    output["result"] = codecs.encode(
                        pickle.dumps(result_from(lattice_record)), "base64"
                    ).decode()
                return jsonify(output)

        except (FileNotFoundError, EOFError):
            if wait:
                continue
            response = make_response(
                jsonify(
                    {
                        "message": "Result not ready to read yet. Please wait for a couple of seconds."
                    }
                ),
                503,
            )
            response.retry_after = 2
            return response
